from enum import Enum

from garpix_utils.logs.enums.get_enums import ActionType, ActionLevel
from garpix_utils.logs.services.action_element import ActionElement


class ActionDefault(Enum):
    # 100-199 | Управление учетными записями пользователей
    user_registration = ActionElement(100, ActionType.user_account, 'Регистрация', ActionLevel.info)
    user_create = ActionElement(101, ActionType.user_account, 'Создание пользователя', ActionLevel.info)
    user_change = ActionElement(102, ActionType.user_account, 'Изменение параметров пользователя', ActionLevel.warning)
    user_delete = ActionElement(103, ActionType.user_account, 'Удаление пользователя', ActionLevel.warning)

    # 200-299 | Идентификация и аутентификация субъекта доступа
    user_login = ActionElement(200, ActionType.authentication, 'Вход в систему', ActionLevel.info)
    user_logout = ActionElement(201, ActionType.authentication, 'Выход из системы', ActionLevel.info)

    # 300-399 | Управление атрибутами доступа
    user_access = ActionElement(300, ActionType.user_access_attribute, 'Изменение привилегий пользователя',
                                ActionLevel.warning)
    group_create = ActionElement(301, ActionType.user_access_attribute, 'Создание группы пользователей',
                                 ActionLevel.warning)
    group_change = ActionElement(302, ActionType.user_access_attribute, 'Изменение группы пользователей',
                                 ActionLevel.warning)
    group_delete = ActionElement(303, ActionType.user_access_attribute, 'Удаление группы пользователей',
                                 ActionLevel.warning)
    group_add_user = ActionElement(304, ActionType.user_access_attribute,
                                   'Занесение учетной записи в группу пользователей', ActionLevel.warning)
    group_delete_user = ActionElement(305, ActionType.user_access_attribute,
                                      'Удаление учетной записи из группы пользователей', ActionLevel.warning)

    # 400-499 | Доступ к защищаемой информации
    user_exist = ActionElement(400, ActionType.access_information, 'Проверка существования пользователя',
                               ActionLevel.info)

    # 500-599 | Удаление информации
    any_entity_delete = ActionElement(500, ActionType.delete_information, 'Удаление сущностей', ActionLevel.info)

    # 600-699 | Добавление информации
    any_entity_create = ActionElement(600, ActionType.create_information, 'Создание сущностей', ActionLevel.info)

    # 700-799 | Изменение информации
    any_entity_change = ActionElement(700, ActionType.change_information, 'Изменение сущностей', ActionLevel.info)

    # 800-899 | Изменение конфигураций
    configuration_change = ActionElement(800, ActionType.change_configuration, 'Изменение конфигураций',
                                         ActionLevel.info)

    '''
        100-199 | Управление учетными записями пользователей
        200-299 | Идентификация и аутентификация субъекта доступа
        300-399 | Управление атрибутами доступа
        400-499 | Доступ к защищаемой информации
        500-599 | Удаление информации
        600-699 | Добавление информации
        700-799 | Изменение информации
        800-899 | Изменение конфигураций
    '''



import uuslug
from garpix_utils.string import get_random_string
import datetime
import os
from django.conf import settings


def get_secret_path(filename=None):
    today = datetime.date.today()
    subdir1 = get_random_string(16)
    subdir2 = get_random_string(16)
    subdir3 = get_random_string(16)
    path = 'data/%s/%s/%s/%s/%s' % (today.year, today.month, subdir1, subdir2, subdir3)
    os.makedirs(os.path.join(settings.MEDIA_ROOT, path))
    if filename is None:
        return path

    ext = filename.split('.')[-1]
    filename = "%s.%s" % (uuslug.slugify(".".join(filename.split('.')[:-1])), ext)

    return os.path.join(path, filename)



import hashlib
from collections import OrderedDict


def make_str(variable):
    """
    Превращает рекурсивно словарь (или другую переменную) в строку
    :param variable:
    :return:
    """
    s = ''
    if type(variable) in (dict, OrderedDict):
        for k, v in variable.items():
            s += '{k}{v}'.format(
                k=k,
                v=make_str(v),
            )
    elif type(variable) in (list, tuple):
        for v in variable:
            s += make_str(v)
    else:
        s += str(variable)
    return s


def recursive_sort(params):
    """
    Рекурсивно сортирует словарь или список.
    :param params:
    :return:
    """
    if type(params) == dict:
        sorted_params = OrderedDict(sorted(params.items(), key=lambda t: t[0]))
        for k, v in sorted_params.items():
            sorted_params[k] = recursive_sort(v)
        return sorted_params
    elif type(params) in (list, tuple):
        params = list(params)
        for index, v in enumerate(params):
            params[index] = recursive_sort(v)
        return sorted(params)
    return params


def make_signature_sha512(params, signature_key='sig', secret='secret'):
    """
    Создает сигнатуру по указанным параметрам с хэшированием SHA-512.
    Алгоритм:
    1. Берет словарь параметров, удаляет оттуда параметр с ключом сигнатуры (см. переменную signature_key, по умолчанию значение "sig")
    2. Получившийся словарь сортирует по названию ключа в алфавитном порядке. Все вложенные данные тоже сортируются по ключу, списочные - просто по алфавиту.
    3. Последовательно конкатенирует ключ со значением в единую строку.
    4. В конце конкатенирует значение переменной secret (по умолчанию равна "secret").
    5. Хэширует по алгоритму SHA-512 и возвращает строку в нижнем регистре.
    6. Возвращает получившийся результат.
    ПРИМЕР:
    1. Было: {'a': 'xxx', 'c': 'ggg', 'b': '111', 'sig': '123', 'd': [3, 1, 2], 'e': {'b': '2', 'a': '1'}}
    Стало: {'a': 'xxx', 'c': 'ggg', 'b': '111', 'd': [3, 1, 2], 'e': {'b': '2', 'a': '1'}}
    2. Было: {'a': 'xxx', 'c': 'ggg', 'b': '111', 'd': [3, 1, 2], 'e': {'b': '2', 'a': '1'}}
    Стало: {'a': 'xxx', 'b': '111', 'c': 'ggg', 'd': [1, 2, 3], 'e': {'a': '1', 'b': '2'}}
    3. Было: {'a': 'xxx', 'b': '111', 'c': 'ggg', 'd': [1, 2, 3], 'e': {'a': '1', 'b': '2'}}
    Стало: 'axxxb111cgggd123ea1b2'
    4. Было: 'axxxb111cgggd123ea1b2'
    Стало: 'axxxb111cgggd123ea1b2secret'
    5. Было: 'axxxb111cgggd123ea1b2secret'
    Стало: '2123086085ec1fe67595d7b3d2b6a0dbf3f33e528d78366b8d62d7f0a7e3c090077b0f7b8dc84921a6087aa57b8284bd1e74702df7a16e96f73f627e6eea815a'
    :param params: dict - словарь параметров. Если присутствует signature_key, то он будет удален.
    :param signature_key: str - ключ параметра с сигнатурой. По умолчанию "sig".
    :param secret: str - секретный ключ, который будет приконкатенирован в конце перед хэшированием.
    :return:
    """
    if signature_key in params:
        del params[signature_key]
    s = make_str(recursive_sort(params)) + secret
    return str(hashlib.sha512(s.encode('utf-8')).hexdigest()).lower()



import socket
from datetime import datetime, timezone

from django.conf import settings
import logging

from garpix_utils.logs.enums.get_enums import ActionResult
from garpix_utils.logs.services.action_element import ActionElement


class LoggerIso:

    def __init__(self, logger_name=settings.ISO_LOGS_NAME):
        self.logger = logging.getLogger(logger_name)

    def create_log(self, action: ActionElement, obj, obj_address, result: ActionResult,
                   params=None, sbj=None, sbj_address=None, msg=''):

        dt = datetime.now(timezone.utc).astimezone()

        log = f'time = {dt.isoformat(timespec="seconds")} | id={action.id} | act=\"{action.name}\"'
        log += f' | sbj=\"{sbj}\"' if sbj else ''
        log += f' | sbj_addr=\"{sbj_address}\"' \
               f' | act_type=\"{action.type.value}\"' \
               f' | lvl=\"{action.level.value}\" | obj=\"{obj}\"' \
               f' | obj_addr=\"{obj_address}\"' \
               f' | result=\"{result.value}\"'
        log += f' | change=\"{params}\"' if params else ''
        log += f' | msg=\"{msg}\"'
        hostname, local_ip = self.get_host_info()
        log += f' | host=\"{hostname}\"' \
               f' | host_addr=\"{local_ip}\"'\
               f' | vendor=\"{getattr(settings, "ISO_LOGS_VENDOR", "Garpix")}\"'\
               f' | product=\"{settings.ISO_LOGS_PRODUCT}\"'
        return log

    def write(self, action: ActionElement, obj, obj_address, result: ActionResult, params=None, sbj=None, sbj_address=None, msg=""):
        log = self.create_log(action, obj, obj_address, result, params, sbj, sbj_address, msg)
        self.logger.info(log)

    def write_string(self, string):
        self.logger.info(string)

    @staticmethod
    def get_client_ip(request):
        # Используется для поулчения адреса субъекта
        x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
        if x_forwarded_for:
            ip = x_forwarded_for.split(',')[0]
        else:
            ip = request.META.get('REMOTE_ADDR')
        return ip

    @staticmethod
    def get_host_info() -> (str, str):
        hostname = socket.gethostname()
        local_ip = socket.gethostbyname(hostname)
        return hostname, local_ip



from garpix_utils.logs.enums.get_enums import ActionType, ActionLevel


class ActionElement:

    def __init__(self, action_id: int, action_type: ActionType, name: str, level: ActionLevel):
        self.id = action_id
        self.name = name
        self.type = action_type
        self.level = level



from django.contrib.sites.models import Site
from django.db import models
from django.db.models import Manager

from garpix_utils.managers import GCurrentSiteManager
from django.core.cache import cache


class GarpixSiteConfiguration(models.Model):
    cache_name = 'garpix_site_config'

    site = models.ForeignKey(Site, on_delete=models.CASCADE)

    objects = Manager()
    on_site = GCurrentSiteManager()

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        self.set_to_cache()

    def delete(self, *args, **kwargs):
        self.clear_cache()
        super().delete(*args, **kwargs)

    @classmethod
    def clear_cache(cls):
        cache.delete(cls.get_cache_key())

    def set_to_cache(self):
        cache.set(self.get_cache_key(), self)

    @classmethod
    def get_cache_key(cls):
        current_site = Site.objects.get_current()
        return f'{cls.cache_name}_{current_site.pk}'

    @classmethod
    def get_solo(cls):

        obj = cache.get(cls.get_cache_key())
        if not obj:
            current_site = Site.objects.get_current()
            obj = cls.objects.filter(site=current_site).first() or cls.objects.create(site=current_site)
            obj.set_to_cache()
        return obj

    def __str__(self):
        return f'Настройки для сайта {self.site.name}'

    class Meta:
        verbose_name = 'Настройки сайта'
        verbose_name_plural = 'Настройки сайтов'
        abstract = True



from garpix_utils.logs.enums.get_enums import Action
from garpix_utils.logs.services.logger_iso import LoggerIso, ActionResult


class CreateLogMixin:
    log_msg_delete = None
    log_msg_create = None
    log_msg_change = None

    @staticmethod
    def log_delete(logger, request, obj, action):

        title = str(obj)

        msg = CreateLogMixin.log_msg_delete if CreateLogMixin.log_msg_delete else f'Объект {title}(id={obj.pk}) модели {obj.__class__.__name__}  был удален'
        return logger.create_log(action=action,
                                 obj=obj.__class__.__name__,
                                 obj_address=request.path,
                                 result=ActionResult.success,
                                 sbj=request.user.username,
                                 sbj_address=LoggerIso.get_client_ip(request),
                                 msg=msg)

    @staticmethod
    def log_change_or_create(logger, request, obj, change, action_change=None, action_create=None):

        title = str(obj)

        if change:
            old_obj = obj.__class__.objects.get(pk=obj.pk) if change else None
            changed_fields = ''
            fields = obj._meta.fields
            for field in fields:
                field_name = field.name
                if hasattr(obj, field_name):
                    old_value = getattr(old_obj, field_name)
                    new_value = getattr(obj, field_name)
                    if old_obj and old_value != new_value:
                        changed_fields += f'{field.verbose_name}: {old_value} -> {new_value};'

            params = changed_fields
            action = action_change or Action.any_entity_change.value
            msg = CreateLogMixin.log_msg_change if CreateLogMixin.log_msg_change else f'Объект {title}(id={obj.pk}) модели {obj.__class__.__name__} был изменен'
        else:
            action = action_create or Action.any_entity_create.value
            msg = CreateLogMixin.log_msg_create if CreateLogMixin.log_msg_create else f'Объект {title} модели {obj.__class__.__name__} был добавлен'
            params = None

        return logger.create_log(action=action,
                                 obj=obj.__class__.__name__,
                                 obj_address=request.path,
                                 result=ActionResult.success,
                                 sbj=request.user.username,
                                 params=params,
                                 sbj_address=LoggerIso.get_client_ip(request),
                                 msg=msg)
    @staticmethod
    def log_change_m2m_field(logger, request, admin_obj_super, form, formsets, change, action_change=None, exclude_fields=[]):

        _obj = form.instance
        obj = _obj.__class__.objects.get(pk=_obj.pk) if change else None

        old_obj_fields = {}

        changed_fields = ''
        for field in obj._meta.get_fields():
            field_name = field.name
            if field.many_to_many and hasattr(obj, field_name):
                old_object_manager = getattr(obj, field_name)
                old_obj_fields.update({
                    field_name: set([str(_obj) for _obj in list(old_object_manager.all())])
                })

        admin_obj_super.save_related(request, form, formsets, change)
        for field in obj._meta.get_fields():
            field_name = field.name
            if field.many_to_many and hasattr(obj, field_name) and field_name not in exclude_fields:
                new_object_manager = getattr(obj, field_name)
                new_relations = set([str(_obj) for _obj in list(new_object_manager.all())])

                new_old = new_relations - old_obj_fields[field_name]
                old_new = old_obj_fields[field_name] - new_relations

                if new_old or old_new:
                    changed_fields += f'{field.field.verbose_name}: '
                    if new_old:
                        changed_fields += f'+добавлены {new_old} '
                    if old_new:
                        changed_fields += f'-удалены {old_new} '

        params = changed_fields
        action = action_change or Action.any_entity_change.value

        if params != '':
            msg = CreateLogMixin.log_msg_change if CreateLogMixin.log_msg_change else f'Объект {str(obj)}(id={obj.pk}) модели {obj.__class__.__name__} был изменен'

            return logger.create_log(action=action,
                                     obj=obj.__class__.__name__,
                                     obj_address=request.path,
                                     result=ActionResult.success,
                                     sbj=request.user.username,
                                     params=params,
                                     sbj_address=LoggerIso.get_client_ip(request),
                                     msg=msg)



from django.db import models
from garpix_utils.file.file_field import UploadTo, secret_file_storage
from garpix_utils.string import get_uuid4_hash


class SecretFileMixin(models.Model):
    file = models.FileField(verbose_name='Файл', upload_to=UploadTo('file'), storage=secret_file_storage)
    share_hash = models.CharField(max_length=32, null=True, default=get_uuid4_hash, blank=True)

    class Meta:
        abstract = True



from garpix_utils.logs.loggers import ib_logger
from garpix_utils.logs.services.logger_iso import LoggerIso, ActionResult
from garpix_utils.logs.enums.get_enums import Action


class LogAdminSolo:

    @staticmethod
    def log_change_or_create(request, obj, change):
        old_obj = obj.__class__.objects.get(pk=obj.pk) if change else None
        changed_fields = ''
        fields = obj._meta.fields
        for field in fields:
            field_name = field.name
            old_value = getattr(old_obj, field_name)
            new_value = getattr(obj, field_name)
            if old_obj and old_value != new_value:
                changed_fields += f'{field.verbose_name}: {old_value} -> {new_value};'
        params = changed_fields
        action = Action.configuration_change.value
        msg = f'{obj.__class__._meta.verbose_name} был изменен'

        return ib_logger.create_log(action=action,
                                    obj=obj.__class__.__name__,
                                    obj_address=request.path,
                                    result=ActionResult.success,
                                    sbj=request.user.username,
                                    params=params,
                                    sbj_address=LoggerIso.get_client_ip(request),
                                    msg=msg)

    def save_model(self, request, obj, form, change):
        log = self.log_change_or_create(request, obj, change)
        super().save_model(request, obj, form, change)
        ib_logger.write_string(log)



from django.conf import settings
from django.utils.module_loading import import_string

ActionType = import_string(
    getattr(settings, 'GARPIX_LOG_ACTION_TYPE', 'garpix_utils.logs.enums.action_type.ActionTypeDefault'))
ActionResult = import_string(
    getattr(settings, 'GARPIX_LOG_ACTION_RESULT', 'garpix_utils.logs.enums.action_result.ActionResultDefault'))
ActionLevel = import_string(
    getattr(settings, 'GARPIX_LOG_ACTION_LEVEL', 'garpix_utils.logs.enums.action_level.ActionLevelDefault'))
Action = import_string(getattr(settings, 'GARPIX_LOG_ACTION', 'garpix_utils.logs.enums.action.ActionDefault'))



from django.conf import settings
from garpix_utils.logs.services.logger_iso import LoggerIso

ib_logger = LoggerIso(settings.IB_ISO_LOGS_NAME)
system_logger = LoggerIso(settings.SYSTEM_ISO_LOGS_NAME)



from enum import Enum


class ActionTypeDefault(Enum):
    authentication = "Идентификация и аутентификация субъекта доступа"
    user_account = "Управление учетными записями пользователей"
    user_access_attribute = "Управление атрибутами доступа"
    access_information = "Доступ к защищаемой информации"
    delete_information = "Удаление информации"
    create_information = "Добавление информации"
    change_information = "Изменение информации"
    change_configuration = "Изменение конфигураций"



from django.contrib import admin
from django.contrib.auth import get_user_model

from garpix_utils.logs.enums.get_enums import Action
from garpix_utils.logs.loggers import ib_logger
from garpix_utils.logs.mixins.create_log import CreateLogMixin


User = get_user_model()


class LogAdminMixin(admin.ModelAdmin, CreateLogMixin):
    class Meta:
        abstract = True

    def save_model(self, request, obj, form, change):
        log = self.log_change_or_create(ib_logger, request, obj, change)
        super().save_model(request, obj, form, change)
        ib_logger.write_string(log)

    def save_related(self, request, form, formsets, change):
        if change:
            log = self.log_change_m2m_field(ib_logger, request, super(), form, formsets, change,
                                            action_change=Action.any_entity_change.value)
            ib_logger.write_string(log)
        else:
            super().save_related(request, form, formsets, change)

    def delete_model(self, request, obj):
        action = Action.user_delete.value if obj.__class__ == User else Action.any_entity_delete.value
        log = self.log_delete(ib_logger, request, obj, action)
        super().delete_model(request, obj)
        ib_logger.write_string(log)

    def delete_queryset(self, request, queryset):
        action = Action.user_delete.value if queryset.first().__class__ == User else Action.any_entity_delete.value
        logs = []
        for obj in queryset:
            logs.append(self.log_delete(ib_logger, request, obj, action))
        super().delete_queryset(request, queryset)
        for log in logs:
            ib_logger.write_string(log)



from enum import Enum


def extend_enum(inherited_enum):
    def wrapper(added_enum):
        joined = {}
        for item in inherited_enum:
            joined[item.name] = item.value
        for item in added_enum:
            joined[item.name] = item.value
        return Enum(added_enum.__name__, joined)
    return wrapper



import datetime
import os

import uuslug
from django.conf import settings
from django.core.files.storage import FileSystemStorage


def get_file_path(instance, filename):
    """
    Формирует путь файла относительно года и месяца, чтобы множество файлов не скапливались на одном уровне.
    """
    ext = filename.split('.')[-1]
    today = datetime.date.today()
    filename = f'{uuslug.slugify(".".join(filename.split(".")[:-1]))}.{ext}'
    return f'uploads/{today.year}/{today.month}/{filename}'


class UploadTo:

    def __init__(self, field_name):
        self.field_name = field_name

    def __call__(self, instance, filename):
        """
        Формирует путь файла относительно названия модели, поля, года и месяца, чтобы множество файлов
        не скапливались на одном уровне.
        """
        ext = filename.split('.')[-1]
        today = datetime.date.today()
        filename = f'{uuslug.slugify(".".join(filename.split(".")[:-1]))}.{ext}'
        return f'uploads/{instance.__class__.__name__}/{self.field_name}/{today.year}/{today.month}/{filename}'

    def deconstruct(self):
        return ('garpix_utils.file.UploadTo', [self.field_name], {})


def secret_file_storage():
    fs = FileSystemStorage(location=getattr(settings, 'SECRET_MEDIA_ROOT',
                                            os.path.join(settings.BASE_DIR, '..', 'secret_public', 'media')))
    return fs
